"""
usage: spetlr-test-job submit [-h] [--dry-run] [--wheels WHEELS] --tests TESTS [--task TASK] [--tasks-from TASKS_FROM] (--cluster CLUSTER | --cluster-file CLUSTER_FILE)
                              [--sparklibs SPARKLIBS | --sparklibs-file SPARKLIBS_FILE] [--requirement REQUIREMENT | --requirements-file REQUIREMENTS_FILE] [--main-script MAIN_SCRIPT] [--pytest-args PYTEST_ARGS]
                              [--out-json OUT_JSON] [--wait-for-job]

Run Test Cases on databricks cluster.

optional arguments:
  -h, --help            show this help message and exit
  --dry-run             Don't do anything, only report
  --wheels WHEELS       The glob paths of all wheels under test.
  --tests TESTS         Location of the tests folder. Will be sent to databricks as a whole.
  --task TASK           Single Test file or folder to execute.
  --tasks-from TASKS_FROM
                        path in test archive where each subfolder becomes a task.
  --cluster CLUSTER     JSON document describing the cluster setup.
  --cluster-file CLUSTER_FILE
                        File with JSON document describing the cluster setup.
  --sparklibs SPARKLIBS
                        JSON document describing the spark dependencies.
  --sparklibs-file SPARKLIBS_FILE
                        File with JSON document describing the spark dependencies.
  --requirement REQUIREMENT
                        a python dependency, specified like for pip
  --requirements-file REQUIREMENTS_FILE
                        File with python dependencies, specified like for pip
  --main-script MAIN_SCRIPT
                        Your own test_main.py script file, to add custom functionality.
  --pytest-args PYTEST_ARGS
                        Additional arguments to pass to pytest in each test job.
  --out-json OUT_JSON   File to store the RunID for future queries.
  --upload-to {workspace,dbfs}
                        Where to upload test job files.
  --wait-for-job        After submission, wait for result using cli v2.


"""

import argparse
import inspect
import json
import re
import shutil
import subprocess
import tempfile
from pathlib import Path, PosixPath
from typing import Dict, List, Union
from typing.io import IO

from spetlrtools.test_job import test_main
from spetlrtools.test_job.dbcli import DbCli
from spetlrtools.test_job.RemoteLocation import (
    DbfsLocation,
    RemoteLocation,
    StageArea,
    WorkspaceLocation,
)


# Custom action to handle the deprecation warning
class DeprecatedAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(
            f"WARNING: {option_string} is deprecated and will be removed in future versions."
        )


def setup_submit_parser(subparsers):
    """
    Adds a subparser for the command 'submit'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """

    parser: argparse.ArgumentParser = subparsers.add_parser(
        "submit", description="Run Test Cases on databricks cluster."
    )
    parser.set_defaults(func=submit_main)

    parser.add_argument(
        "--dry-run", help="Don't do anything, only report", action="store_true"
    )

    parser.add_argument(
        "--wheels",
        type=str,
        required=False,
        help="The glob paths of all wheels under test.",
        default="dist/*.whl",
    )

    parser.add_argument(
        "--tests",
        type=str,
        required=True,
        help="Location of the tests folder. Will be sent to databricks as a whole.",
    )

    parser.add_argument(
        "--task", help="Single Test file or folder to execute.", action="append"
    )
    parser.add_argument(
        "--tasks-from",
        help="path in test archive where each subfolder becomes a task.",
        action="append",
    )

    # cluster argument pair
    cluster = parser.add_mutually_exclusive_group(required=True)
    cluster.add_argument(
        "--cluster",
        type=str,
        help="JSON document describing the cluster setup.",
    )
    cluster.add_argument(
        "--cluster-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the cluster setup.",
    )

    # spark libraries argument pair
    sparklibs = parser.add_mutually_exclusive_group(required=False)
    sparklibs.add_argument(
        "--sparklibs",
        help="JSON document describing the spark dependencies.",
    )
    sparklibs.add_argument(
        "--sparklibs-file",
        type=argparse.FileType("r"),
        help="File with JSON document describing the spark dependencies.",
    )

    # python dependencies file
    pydep = parser.add_mutually_exclusive_group(required=False)
    pydep.add_argument(
        "--requirement",
        action="append",
        help="a python dependency, specified like for pip",
        default=[],
    )
    pydep.add_argument(
        "--requirements-file",
        type=argparse.FileType("r"),
        help="File with python dependencies, specified like for pip",
    )

    parser.add_argument(
        "--main-script",
        type=argparse.FileType("r"),
        help="Your own test_main.py script file, to add custom functionality.",
    )

    parser.add_argument(
        "--pytest-args", help="Additional arguments to pass to pytest in each test job."
    )

    parser.add_argument(
        "--out-json",
        type=argparse.FileType("w"),
        help="File to store the RunID for future queries.",
    )

    parser.add_argument(
        "--upload-to",
        choices=["workspace", "dbfs"],
        help="Where to upload test job files.",
        default="dbfs",
    )

    parser.add_argument(
        "--wait-for-job",
        action="store_true",
        help="After submission, wait for result using cli v2.",
    )
    parser.add_argument("--wait", action=DeprecatedAction, help=argparse.SUPPRESS)

    return


def collect_arguments(args):
    """
    Post process the parsed arguments of the 'submit' command argument parser.
    :param args: parsed arguments of the 'submit' command argument parser
    :return:
    """

    # pre-process 'cluster'
    if args.cluster_file:
        args.cluster = args.cluster_file.read()
    args.cluster = json.loads(args.cluster)

    # pre-process 'sparklibs'
    if args.sparklibs_file:
        args.sparklibs = args.sparklibs_file.read()
    if args.sparklibs:
        args.sparklibs = json.loads(args.sparklibs)

    # pre-process 'requirement'
    if args.requirements_file:
        args.requirement = [
            line.strip()
            for line in args.requirements_file.read().splitlines()
            if line.strip() and not line.strip().startswith("#")
        ]

    args.pytest_args = (args.pytest_args or "").split()

    return args


def submit_main(args):
    """the main function of the cli command 'submit'. Not to be used directly."""

    args = collect_arguments(args)

    submit(
        test_path=args.tests,
        cluster=args.cluster,
        wheels=args.wheels,
        tasks=args.task,
        tasks_from=args.tasks_from,
        requirement=args.requirement,
        sparklibs=args.sparklibs,
        out_json=args.out_json,
        main_script=args.main_script,
        pytest_args=args.pytest_args,
        dry_run=args.dry_run,
        upload_to=args.upload_to,
        wait_for_job=args.wait_for_job,
    )


def verify_and_resolve_task(test_path: str, task: Union[str, PosixPath]):
    test_archive = PosixPath(test_path).resolve().absolute()

    task_path = PosixPath(task)
    if not task_path.is_absolute():
        task_path = (test_archive.parent / task_path).resolve().absolute()

    task_path = task_path.resolve().absolute()

    if not (test_archive == task_path or test_archive in task_path.parents):
        raise AssertionError(f"task {task} is not contained in {test_path}")

    if not task_path.exists():
        raise AssertionError(f"task {task_path} does not exist")

    if not (test_archive.parent / task).exists():
        raise AssertionError(f"The task {task} was not found in the test location.")

    return task_path.relative_to(test_archive.parent).as_posix()


def discover_job_tasks(test_path: str, folder: str):
    """If folder is given, create parallel tasks from this level.
    Otherwise, simply return the top folder as the single task.
    Returns a list of strings with the subfolders to process.
    """

    test_archive_parent = PosixPath(test_path).resolve().absolute().parent

    subfolders = [
        verify_and_resolve_task(test_path, x)
        for x in (test_archive_parent / folder).iterdir()
        if (x.is_dir() and not x.stem.startswith("_"))
    ]
    return subfolders


class PoolBoy:
    """Hold a list of available instance pools and replace the by-name reference with an id if possible."""

    MARKER = "instance-pool://"

    def __init__(self):
        self._lookup = self.get_instance_pools()

    def lookup(self, pool_id: str) -> str:
        if pool_id.startswith(self.MARKER):
            pool_name = pool_id[len(self.MARKER) :]
            return self._lookup[pool_name]
            # if this throws a KeyError, we are right to abort execution since the input is invalid.
        else:
            # No Marker = no lookup
            return pool_id

    def get_instance_pools(self) -> Dict[str, str]:
        pool_lookup = {
            pool.instance_pool_name: pool.instance_pool_id
            for pool in DbCli().list_instance_pools()
        }
        return pool_lookup


def submit(
    test_path: str,
    cluster: dict,
    wheels: str,
    tasks: List[str] = None,
    tasks_from: List[str] = None,
    requirement: List[str] = None,
    sparklibs: List[dict] = None,
    out_json: IO[str] = None,
    main_script: IO[str] = None,
    pytest_args: List[str] = None,
    dry_run=False,
    upload_to="dbfs",
    wait_for_job=False,
):
    """
    --dry-run             Don't do anything, only report
    --wheels WHEELS       The glob paths of all wheels under test.
    --tests TESTS         Location of the tests folder. Will be sent to databricks as a whole.
    --task TASK           Single Test file or folder to execute.
    --tasks-from TASKS_FROM
                          path in test archive where each subfolder becomes a task.
    --cluster CLUSTER     JSON document describing the cluster setup.
    --cluster-file CLUSTER_FILE
                          File with JSON document describing the cluster setup.
    --sparklibs SPARKLIBS
                          JSON document describing the spark dependencies.
    --sparklibs-file SPARKLIBS_FILE
                          File with JSON document describing the spark dependencies.
    --requirement REQUIREMENT
                          a python dependency, specified like for pip
    --requirements-file REQUIREMENTS_FILE
                          File with python dependencies, specified like for pip
    --main-script MAIN_SCRIPT
                          Your own test_main.py script file, to add custom functionality.
    --pytest-args PYTEST_ARGS
                          Additional arguments to pass to pytest in each test job.
    --out-json OUT_JSON   File to store the RunID for future queries.
    --upload-to {workspace,dbfs}
                          Where to upload test job files.
    --wait-for-job        After submission, wait for result using cli v2.
    """
    if requirement is None:
        requirement = []
    if sparklibs is None:
        sparklibs = []
    if pytest_args is None:
        pytest_args = []
    if tasks is None:
        tasks = []
    if tasks_from is None:
        tasks_from = []
    if not (tasks or tasks_from):
        raise ValueError("No tasks given")
    upload_to = upload_to.lower()

    # check the structure of the cluster object
    if not isinstance(cluster, dict):
        raise AssertionError("invalid cluster specification")

    # check the structure of the sparklibs object
    if not isinstance(sparklibs, list):
        raise AssertionError("invalid sparklibs specification")

    for py_requirement in requirement:
        sparklibs.append({"pypi": {"package": py_requirement}})

    dbcli = DbCli()

    # create everything in a temporary directory.
    # for dry-runs, keep make it a local directory and keep it
    with StageArea(dry_run) as stage:
        if upload_to == "workspace":
            remote: RemoteLocation = WorkspaceLocation(stage)
        elif upload_to == "dbfs":
            remote: RemoteLocation = DbfsLocation(stage)
        else:
            raise ValueError("unsupported upload")

        wheels = discover_wheels(wheels, remote)
        for wheel in wheels:
            sparklibs.append({"whl": wheel})

        prepare_archive(test_path, remote)
        main_file = prepare_main_file(remote, main_script)

        resolved_tasks = [verify_and_resolve_task(test_path, task) for task in tasks]
        for task in tasks_from:
            # subtasks will be ['tests/cluster/job1', 'tests/cluster/job2'] or similar
            resolved_tasks += discover_job_tasks(test_path, task)

        if dry_run:
            print(resolved_tasks)

        if "instance_pool_id" in cluster:
            cluster["instance_pool_id"] = PoolBoy().lookup(cluster["instance_pool_id"])

        # construct the workflow object
        workflow = dict(run_name="Testing Run", format="MULTI_TASK", tasks=[])

        for task in resolved_tasks:
            # construct a task name from the test task file path
            task_sub = re.sub(r"[^a-zA-Z0-9_-]", "_", task)

            workflow["tasks"].append(
                dict(
                    task_key=task_sub,
                    libraries=sparklibs,
                    spark_python_task=dict(
                        python_file=main_file,
                        parameters=[
                            # running in the spark python interpreter, the python __file__ variable does not
                            # work. Hence, we need to tell the script where the test area is.
                            f"--basedir={remote.remote_base()}",
                            # we can actually run any part of our test suite, but some files need the full repo.
                            # Only run tests from this folder.
                            f"--folder={task}",
                            # additional arguments to pass to pytest
                            f"--pytestargs={json.dumps(pytest_args)}",
                        ],
                    ),
                    new_cluster=cluster,
                )
            )

        jobfile = remote.new_local_file("job.json").local

        with open(jobfile, "w") as f:
            json.dump(workflow, f, indent=2)

        remote.upload(dry_run)

        if wait_for_job:
            print("handing control to databricks jobs submit ...")
            dbcli.execv_run_file(jobfile, dry_run=dry_run)
            # the above function ends python and does not return
            return

        try:
            print("Submitting job...")
            run_id = dbcli.submit(workflow, dry_run=dry_run)
        except subprocess.CalledProcessError:
            print("Json contents:")
            print(json.dumps(workflow, indent=4))
            raise

    # now we have the run_id
    print(f"Started run with ID {run_id}")
    print(f"Follow job details at {dbcli.get_run(run_id).run_page_url}")

    if out_json:
        json.dump({"run_id": run_id}, out_json)


def discover_wheels(globpath: str, remote: RemoteLocation) -> List[str]:
    """Find all wheel files in the globpath and add them to the remote location."""
    result = []
    for item in Path().glob(globpath):
        result.append(remote.add_local_path(str(item), "libs"))

    return result


def prepare_archive(test_path: str, remote: RemoteLocation):
    """Zip the test archive and add it to the staging area"""
    print(f"now archiving {test_path}")

    with tempfile.TemporaryDirectory() as tempdir:
        real_archive_path = shutil.make_archive(
            str(Path(tempdir) / "tests"),
            "zip",
            Path(test_path) / "..",
            base_dir=Path(test_path).parts[-1],
        )

        # it seems the doing a workspace import-dir on a zip archive will unpack it locally to upload.
        # so we need to trick it by renaming the file
        renamed_archive_path = Path(real_archive_path).with_suffix(".archive")
        shutil.move(real_archive_path, renamed_archive_path)

        return remote.add_local_path(str(renamed_archive_path))


def prepare_main_file(remote: RemoteLocation, main_script: IO[str] = None) -> str:
    print("now preparing test main file")

    main_ref = remote.new_local_file("main.py")

    with open(main_ref.local, "w") as f:
        if main_script:
            f.write(main_script.read())
        else:
            print("Using default main script test_main.py")
            f.write(inspect.getsource(test_main))

    return main_ref.remote



"""
This script provides functionality for managing the installation process of a CLI tool.
It includes commands to fetch the latest version of the CLI. Key functionalities include:
- Adding a subparser for the 'latest' command to check for the latest version of the CLI.
- A main function for the 'latest' command, which fetches and prints the latest CLI version.
- A utility function to retrieve the latest version from a specified URL.

Note: This script is intended to be used as a part of a larger CLI tool and should
 be invoked through the command line interface.
"""

import argparse
import sys

import requests


def setup_latest_parser(subparsers):
    """
    Adds a subparser for the command 'install'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "latest", description="Get the latest cli version."
    )
    parser.set_defaults(func=latest_main)


def latest_main(args):
    """
    Main function of the 'uninstall' command. Only to be used via the cli.
    :param args: the parsed arguments from the install subparser
    :return:
    """

    latest_version = get_latest_version()
    if latest_version:
        print(latest_version)
    else:
        print("Fetching latest version failed")
        sys.exit(-1)


def get_latest_version():
    """Uninstall main function."""
    r = requests.get(
        "https://raw.githubusercontent.com/databricks/setup-cli/main/VERSION"
    )
    return r.content.decode("utf-8").strip()



import sys
from json import JSONDecodeError
from pprint import pprint

import requests


def get_ad_access_token(
    appId: str,
    tenantId: str,
    auth_code: str,
    state: str,
    redirect_uri: str,
    verbose: bool = False,
):
    if verbose:
        print("  get_ad_access_token:", file=sys.stderr)
        print(f"     appId:        {appId[:7]}...", file=sys.stderr)
        print(f"     tenantId:     {tenantId[:7]}...", file=sys.stderr)
        print(f"     auth_code:    {auth_code[:7]}...", file=sys.stderr)
        print(f"     state:        {state[:7]}...", file=sys.stderr)
        print(f"     redirect_uri: {redirect_uri}", file=sys.stderr)

    url = f"https://login.microsoftonline.com/{tenantId}/oauth2/v2.0/token"
    data = dict(
        client_id=appId,
        scope="2ff814a6-3304-4ab8-85cb-cd0e6f879c1d/.default",
        code=auth_code,
        redirect_uri=redirect_uri,
        grant_type="authorization_code",
        state=state,
    )
    r = requests.post(url=url, data=data)

    try:
        result = r.json()
    except JSONDecodeError:
        raise Exception(f"Unexpected result from login.microsoftonline.com:\n{r.text}")

    if "access_token" not in result:
        if verbose:
            print("Error received:", file=sys.stderr)
            pprint(result, stream=sys.stderr)

        if "error_codes" in result:
            raise Exception(f"Login Exception code {result['error_codes'].pop()}")
        else:
            raise Exception("Unknown Login Exception")

    return result["access_token"]



"""
This script acts as the primary entry point for a Databricks-related CLI tool. It orchestrates various commands
including:
- Setting up the CLI environment and parsing command-line arguments.
- Integrating subparsers for 'install', 'uninstall', and 'latest' commands, linking to their respective functionalities.
- Handling the execution flow based on the command chosen by the user.

The script leverages 'argparse' for command-line parsing, providing a user-friendly interface for interacting with the
 Databricks CLI tool. It is designed to be invoked from the command line and not intended for
 use as a module in other Python scripts.

Overall, this script simplifies the user interaction with the Databricks CLI tool,
offering a streamlined way to install, uninstall, and check for updates directly from the command line.
"""

import argparse

from spetlrtools.cli_install.install import setup_install_parser
from spetlrtools.cli_install.latest import setup_latest_parser
from spetlrtools.cli_install.uninstall import setup_uninstall_parser


def main():
    """Cli main function. Not for use from python."""
    parser = argparse.ArgumentParser(
        description="Run Test Cases on databricks cluster."
    )

    subparsers = parser.add_subparsers(required=True, dest="command")

    setup_install_parser(subparsers)
    setup_uninstall_parser(subparsers)
    setup_latest_parser(subparsers)

    args = parser.parse_args()
    args.func(args)



import threading
import time
import uuid
import webbrowser
from urllib.parse import urlencode


class AuthLinkOpener(threading.Thread):
    def __init__(self, appId: str, tenantId: str, local_port: int):
        super().__init__()
        self.appId = appId
        self.tenantId = tenantId
        self.local_port = local_port
        self.state = str(uuid.uuid4())
        self.start()

    def build_query(self):
        url = (
            f"https://login.microsoftonline.com/{self.tenantId}"
            "/oauth2/v2.0/authorize?"
        )
        self.redirect_uri = f"http://localhost:{self.local_port}"
        query = {
            "client_id": self.appId,
            "response_type": "code",
            "redirect_uri": self.redirect_uri,
            "response_mode": "query",
            # fixed scope for Azure Databricks
            "scope": "2ff814a6-3304-4ab8-85cb-cd0e6f879c1d/.default",
            "state": self.state,
        }

        return url + urlencode(query)

    def run(self) -> None:
        time.sleep(2)
        webbrowser.open(self.build_query())



from functools import total_ordering


@total_ordering
class TupleComparer(object):
    """This class contains a potentially nested tuple and can be
    compared even if any of the sub-members are zero.
    The only requirements are that the tuples are of equal length,
    and that they have comparable types in non-zero fields."""

    def __init__(self, data: tuple):
        self.data = data

    def __le__(self, other: "TupleComparer"):
        if self == other:
            return True
        if len(self.data) != len(other.data):
            raise ValueError("Cannot compare tuples of unequal lengths.")

        for lhs, rhs in zip(self.data, other.data):
            if lhs is None:
                if rhs is None:
                    # Equal, check more elements.
                    continue
                else:
                    # None is less than non-none
                    return True
            else:
                # lsh is not none
                if rhs is None:
                    # nothing is less than None
                    return False
                else:
                    # both not None
                    if lhs == rhs:
                        continue

                    if isinstance(lhs, tuple) and isinstance(rhs, tuple):
                        return TupleComparer(lhs) <= TupleComparer(rhs)

                    # neither is None nor tuple and unequal
                    return lhs < rhs

        # if not element comparison could reach a conclusion
        # something broke
        raise ValueError("Inconclusive comparison between unequal tuples")

    def __eq__(self, other):
        return self.data == other.data



"""
This package implements the MSAL flow to obtain a personal api token for an instance
of azure databricks without using the web interface.

see url:
https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/aad/app-aad-token

see also:
https://docs.microsoft.com/en-us/azure/active-directory/develop/reply-url#localhost-exceptions

Required input:
- applicationId (for the auth application)
- tenantId


"""

# from spetlrtools.az_databricks_token.main import main
# from spetlrtools.az_databricks_token.server import LocalRequestHandler
#



import argparse
from pathlib import Path

from spetlrtools.az_databricks_token.get_ad_access_token import get_ad_access_token
from spetlrtools.az_databricks_token.get_impersonation_authorization_code import (
    get_impersonation_authorization_code,
)


def main():
    parser = argparse.ArgumentParser(
        description="Request Azure Databricks MSAL token.",
        epilog="For more information see https://docs.microsoft.com"
        "/en-us/azure/databricks/dev-tools/api/latest/aad/app-aad-token",
    )
    parser.add_argument("--appId", required=True, help="the authorizing webapp")
    parser.add_argument("--tenantId", required=True, help="the tenant of the webapp")
    parser.add_argument(
        "--workspaceUrl", help="directly set .databrickscfg if this is given"
    )
    # help="debug printing"
    parser.add_argument("-v", "--verbose", action="store_true", help=argparse.SUPPRESS)

    args = parser.parse_args()

    # 1. Request an authorization code, which launches a browser window and asks
    # for Azure user login. The authorization code is returned after the user
    # successfully logs in.
    results = get_impersonation_authorization_code(
        appId=args.appId, tenantId=args.tenantId
    )

    # 2. Use the authorization code to acquire the Azure AD access token.
    access_token = get_ad_access_token(
        appId=args.appId,
        tenantId=args.tenantId,
        auth_code=results.auth_code,
        state=results.state,
        redirect_uri=results.redirect_uri,
        verbose=args.verbose,
    )

    if args.workspaceUrl:
        dbcfg = (
            "[DEFAULT]\n"
            f"host = https://{args.workspaceUrl}\n"
            f"token = {access_token}"
        )
        with open(Path.home() / ".databrickscfg", "w") as f:
            f.write(dbcfg)
        print("Your local .databrickscfg has been updated with a token.")
    else:
        print(access_token)



import os
import shutil
import sys
from typing import Any, Iterator, Optional

from databricks.sdk import WorkspaceClient
from databricks.sdk.service import jobs
from databricks.sdk.service.compute import InstancePoolAndStats


def _try_resolve(obj: Any, key: str):
    try:
        return obj[key]
    except (KeyError, TypeError):
        return obj


class DbCli:
    w: Optional[WorkspaceClient] = None

    @classmethod
    def get_client(cls) -> WorkspaceClient:
        if cls.w is None:
            cls.w = WorkspaceClient()
        return cls.w

    def whoami(self) -> str:
        return self.get_client().current_user.me().user_name

    def cancel_run(self, run_id: int) -> None:
        self.get_client().jobs.cancel_run(run_id)

    def get_run(self, run_id: int) -> jobs.Run:
        return self.get_client().jobs.get_run(run_id)

    def get_run_output(self, run_id: int) -> jobs.RunOutput:
        return self.get_client().jobs.get_run_output(run_id)

    def list_instance_pools(self) -> Iterator[InstancePoolAndStats]:
        return self.get_client().instance_pools.list()

    def submit(self, workflow: dict, dry_run=False) -> int:
        if dry_run:
            print("Action skipped for dry-run. Job not submitted.")
            print("You can find the json to be submitted in the staging area.")
            print("Dry run ends here.")
            sys.exit(0)

        return self.get_client().jobs._api.do(
            "POST", "/api/2.1/jobs/runs/submit", body=workflow
        )["run_id"]

    def execv_run_file(self, file_path: str, dry_run=False):
        databricks = shutil.which("databricks")
        if databricks is None:
            print("databricks not found in PATH.", file=sys.stderr)
            sys.exit(1)

        args = ["databricks", "jobs", "submit", f"--json=@{file_path}"]

        if dry_run:
            print("Action skipped for dry-run:")
            print(">>", *args)
            sys.exit(0)

        try:
            os.execv(databricks, args)
        except Exception as e:
            print(f"Failed to execute the command: {e}", file=sys.stderr)
            sys.exit(1)



"""
This script is designed for setting up environment-specific configurations for the Databricks CLI tool.
It contains functionality to:
- Detect the operating system of the user and determine if it is Windows.
- Define default target paths for the Databricks CLI executable based on the operating system.
- Store the GitHub repository URL of the Databricks CLI for reference or further use.

The script uses the 'platform' and 'sysconfig' modules to accurately identify the environment and
set file paths, ensuring compatibility with different operating systems, particularly Windows and Unix-based systems.
"""

import platform
import re
import sysconfig
from pathlib import Path


def isWin():
    return bool(re.match(".*windows.*", platform.system(), flags=re.IGNORECASE))


_DEFAULT_TARGET = Path(sysconfig.get_paths()["scripts"]) / (
    "databricks.exe" if isWin() else "databricks"
)
_DATABRICKS_CLI_REPO = "https://github.com/databricks/cli"



import json

import pyspark.sql.functions as f
import pyspark.sql.types as t
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType


class ExtractEncodedBodyUC:
    """
    Utility class to extract and parse JSON bodies from a binary encoded column
    This can be used for investigating eventhub data.

    Attributes:
        json_field (str): Name of the column containing JSON strings to be parsed.
        new_field_w_extracted_body (str): Name of the column to store parsed JSON as a StructType.
        _data_limit (int or None): Maximum number of rows to sample when inferring JSON schema.
    """

    def __init__(
        self,
        json_field: str = "Body",
        new_column_w_extracted_body: str = None,
        data_limit: int = None,
    ):
        self.json_field = json_field
        self.new_field_w_extracted_body = new_column_w_extracted_body or json_field
        self._data_limit = data_limit
        self._fields_have_same_name = self.json_field == self.new_field_w_extracted_body

    def extract_schema(self, df: DataFrame) -> StructType:
        """
        Infer and return the StructType schema of the JSON column.

        Args:
            df (DataFrame): Input DataFrame.

        Returns:
            StructType: Schema of the parsed JSON field.

        Raises:
            AssertionError: If `json_field` not in DataFrame.
        """
        assert self.json_field in df.columns, "The body column is not in the dataframe"

        total_schema = self.transform_df(df).schema

        return StructType([total_schema[self.new_field_w_extracted_body]])

    def extract_schema_as_json(self, df: DataFrame, pretty_json: bool = False) -> str:
        """
        Return the JSON schema of the parsed field as a string.

        Args:
            df (DataFrame): Input DataFrame.
            pretty_json (bool): If True, format with indentation.

        Returns:
            str: JSON-formatted schema.
        """
        schema = self.extract_schema(df)
        json_schema_raw = schema.jsonValue()
        return json.dumps(json_schema_raw, indent=4 if pretty_json else None)

    def transform_df(
        self, df: DataFrame, keep_original_body: bool = False
    ) -> DataFrame:
        """
        Parse the JSON column into a StructType column.

        Args:
            df (DataFrame): Input DataFrame.
            keep_original_body (bool): If True and names match, error is raised.
                If False and names differ, drops original JSON column.

        Returns:
            DataFrame: DataFrame with parsed JSON column.

        Raises:
            ValueError: If attempting to overwrite the original JSON column.
        """
        if self._fields_have_same_name and keep_original_body:
            raise ValueError(
                f"The field {self.json_field} is overwritten by the extracted json. Give extracted field new name to keep original body."
            )

        df_string = df.withColumn(
            self.new_field_w_extracted_body, f.col(self.json_field).cast(t.StringType())
        )

        sample_jsons = df_string

        if self._data_limit:
            sample_jsons = sample_jsons.limit(self._data_limit)

        sample_jsons = sample_jsons.select(self.new_field_w_extracted_body).agg(
            f.first(self.new_field_w_extracted_body)
        )

        if self._data_limit:
            sample_jsons = sample_jsons.limit(self._data_limit)

        sample_jsons = sample_jsons.collect()[0][0]

        df_parsed = df_string.withColumn(
            self.new_field_w_extracted_body,
            f.from_json(
                f.col(self.new_field_w_extracted_body), f.schema_of_json(sample_jsons)
            ),
        )

        if (not keep_original_body) and (not self._fields_have_same_name):
            df_parsed = df_parsed.drop(self.json_field)

        return df_parsed



import datetime
import unittest
from typing import Any, Iterable

from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, StructField, StructType

from spetlrtools.testing.TupleComparer import TupleComparer


class DataframeTestCase(unittest.TestCase):
    def assertDataframeMatches(
        self,
        df: DataFrame,
        columns: Iterable[str] = None,
        expected_data: Iterable[Iterable[Any]] = None,
        assert_order=False,
    ):
        """
        Args:
            df:
            columns: the column names in the expected data. In case of None, all df column are used.
            expected_data: expected data, for the selected columns.
                    This necessary argument has a default value because columns can be set to default.

        Returns:
            None

        Raises:
            AssertionError in case the dataframe does not match the expected data.
        """
        if expected_data is None:
            raise ValueError("No match assert value provided")
        if columns is None:
            columns = df.schema.fieldNames()

        schema_fields = {f.name: f for f in df.schema.fields}

        assert_df = [list(row) for row in df.select(*columns).collect()]

        try:
            # we need to handle array types
            # this code should grow to handle more general types of data-frames in the future
            for i, col in enumerate(columns):
                if isinstance(schema_fields[col].dataType, ArrayType):
                    for row in assert_df:
                        if row[i] is None:
                            continue
                        elif isinstance(
                            schema_fields[col].dataType.elementType, StructType
                        ):
                            row[i] = [
                                tuple(
                                    (
                                        field.astimezone(datetime.timezone.utc)
                                        if isinstance(field, datetime.datetime)
                                        else field
                                    )
                                    for field in sub_row
                                )
                                for sub_row in row[i]
                            ]

                # move every datetime item to UTC time zone. This allows for exact comparison
                for row in assert_df:
                    if isinstance(row[i], datetime.datetime):
                        row[i] = row[i].astimezone(datetime.timezone.utc)

            assert_df = [tuple(row) for row in assert_df]

            if not assert_order:
                assert_df = sorted(assert_df, key=TupleComparer)

        except KeyError as e:
            raise KeyError(
                f'Could not find "{e.args[0]}" column in dataframe schema: {list(schema_fields)}'
            )

        # move every datetime item to UTC time zone. This allows for exact comparison
        expected_data = [list(row) for row in expected_data]
        for row in expected_data:
            for i, item in enumerate(row):
                if isinstance(row[i], datetime.datetime):
                    row[i] = row[i].astimezone(datetime.timezone.utc)
        expected_data = [tuple(row) for row in expected_data]

        if not assert_order:
            expected_data = sorted(expected_data, key=TupleComparer)

        self.assertEqual(expected_data, assert_df)

    def assertEqualSchema(
        self, schema1: StructType, schema2: StructType, compare_nullability=False
    ):
        """
        Comparing pyspark schemas can fail due to unequal nullability.
        This method allows the user to ignore this aspect of the StructType.
        """
        json1 = schema1.jsonValue()
        json2 = schema2.jsonValue()

        if not compare_nullability:
            # create a function     for recursive clearing
            def clear_nullable(field: StructField):
                del field["nullable"]
                if not field["metadata"]:
                    del field["metadata"]
                if field["type"] == "struct":
                    for item in field["fields"]:
                        clear_nullable(item)
                elif (
                    field["type"] == "array"
                    and field["elementType"]["type"] == "struct"
                ):
                    for item in field["elementType"]["fields"]:
                        clear_nullable(item)

            for j in [json1, json2]:
                for item in j["fields"]:
                    clear_nullable(item)

        self.assertEqual(json1, json2)



import importlib
import pkgutil
import sys
from inspect import getmembers, isclass
from types import ModuleType
from typing import Dict, Union


class ModuleHelper:
    @staticmethod
    def get_modules(package: Union[str, ModuleType]) -> Dict[str, ModuleType]:
        """
        Retrieves all modules from a given package or module.

        Args:
            package (Union[str, ModuleType]):
                The name of the package or a reference to the module.

        Returns:
            Dict[str, ModuleType]: A dictionary containing all modules.
        """
        modules = {}

        package = importlib.import_module(package)

        if not hasattr(package, "__path__"):
            modules[package.__name__] = package

        else:
            for ModuleInfo in pkgutil.walk_packages(package.__path__):
                module_name = f"{package.__name__}.{ModuleInfo.name}"

                if ModuleInfo.ispkg:
                    modules.update(ModuleHelper.get_modules(module_name))

                else:
                    module = importlib.import_module(module_name)
                    modules.update({module_name: module})

        return modules

    @staticmethod
    def get_classes_of_type(
        package: Union[str, ModuleType],
        obj: type,
        main_classes: bool = True,
        sub_classes: bool = True,
    ) -> Dict[str, dict]:
        """
        Retrieves all classes of a specified type from a package or module.

        Args:
            package (Union[str, ModuleType]):
                The name of the package or a reference to the module.
            obj (type):
                The type of classes to retrieve.
            main_classes (bool) default = True:
                If the main classes of the type obj should be retrieved.
            sub_classes (bool) default = True:
                If the sub classes of the type obj should be retrieved.

        Returns:
            Dict[str, dict]:
                A dictionary containing all classes of the specified type,
                along with module and class information.
        """
        modules = ModuleHelper.get_modules(package=package)

        objects = {}
        for module_name, module in modules.items():
            for cls_name, cls in getmembers(sys.modules[module_name], isclass):
                if main_classes:
                    if cls_name == obj.__name__:
                        objects[f"{module_name}.{cls_name}"] = {
                            "module_name": module_name,
                            "module": module,
                            "cls_name": cls_name,
                            "cls": cls,
                        }

                if sub_classes:
                    for base_cls in cls.__bases__:
                        if base_cls.__name__ == obj.__name__:
                            objects[f"{module_name}.{cls_name}"] = {
                                "module_name": module_name,
                                "module": module,
                                "cls_name": cls_name,
                                "cls": cls,
                            }

        return objects



from typing import List

from spetlr.entry_points import TaskEntryPoint

from spetlrtools.helpers import ModuleHelper


class TaskEntryPointHelper:
    @staticmethod
    def get_all_task_entry_points(
        packages: List[str],
        output_txt_file: str = None,
        entry_point_objects: List[type] = None,
    ) -> dict:
        """
        Returns a dictionary of entry points for all `TaskEntryPoint` objects found
        in the specified packages.

        Args:
            packages (List[str]):
                A list of package names to search for `TaskEntryPoint` objects.
            output_txt_file (str, optional) default = None:
                The name of a text file to write the entry points to.
            entry_point_objects (List[type], optional) default = None:
                One or more objects to get entry points from. Use for custom base
                classes that has a `task()` abstract class method.

        Returns:
            dict:
                A dictionary of entry points, with a single key called
                'spetlrtools.task_entry_points'. The value of the key is a list of all
                discovered `TaskEntryPoint` classes with the `task()` method
                implemented.
        """
        if not isinstance(packages, List):
            raise TypeError("The 'packages' argument need to be of type 'List'")

        entry_point_objs = {}

        if entry_point_objects is not None:
            for entry_point_object in entry_point_objects:
                for package in packages:
                    entry_point_objs.update(
                        ModuleHelper.get_classes_of_type(
                            package=package,
                            obj=entry_point_object,
                            main_classes=False,
                            sub_classes=True,
                        )
                    )
        else:
            for package in packages:
                entry_point_objs.update(
                    ModuleHelper.get_classes_of_type(
                        package=package,
                        obj=TaskEntryPoint,
                        main_classes=False,
                        sub_classes=True,
                    )
                )

        entry_points = {"spetlrtools.task_entry_points": []}

        for module_and_cls, contents in entry_point_objs.items():
            entry_point_name = module_and_cls
            module_name = contents["module_name"]
            cls_name = contents["cls_name"]
            entry_points["spetlrtools.task_entry_points"].append(
                f"{entry_point_name} = {module_name}:{cls_name}.task"
            )

        if output_txt_file:
            with open(output_txt_file, "w") as auto_ep_file:
                for entry_point in entry_points["spetlrtools.task_entry_points"]:
                    print(f"Found entry point: {entry_point}")
                    auto_ep_file.write(f"{entry_point}\n")

        return entry_points



"""
This script facilitates the installation of the Databricks CLI tool. Key features include:
- Adding a subparser for the 'install' command to manage CLI installation.
- The main function for the 'install' command, handling installation process based on user-specified
    or default settings.
- Determining the operating system and architecture to download the correct version of the CLI tool.
- Functionality to download and install a specific or the latest version of the Databricks CLI, with options for
    version control and target installation paths.
- Force uninstallation of existing versions if specified.

Dependencies include modules for command-line argument parsing, system and environment management,
and internet requests.
Note: This script is intended to be used through a command line interface and assumes access to required system paths
and internet connectivity.
"""

import argparse
import io
import os
import platform
import re
import subprocess
import sys
import zipfile

import requests

from spetlrtools.cli_install.defaults import (
    _DATABRICKS_CLI_REPO,
    _DEFAULT_TARGET,
    isWin,
)
from spetlrtools.cli_install.latest import get_latest_version
from spetlrtools.cli_install.uninstall import uninstall


def setup_install_parser(subparsers):
    """
    Adds a subparser for the command 'install'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "install", description="Install the datbricks cli v2 to your python area."
    )
    parser.set_defaults(func=install_main)

    parser.add_argument(
        "--version",
        type=str,
        help="Version of CLI to install (accepts @file), defaults to latest.",
        default=None,
    )

    parser.add_argument(
        "--target",
        type=str,
        help="path of the installed cli",
        default=_DEFAULT_TARGET,
    )

    parser.add_argument(
        "--force",
        action="store_true",
        help="Remove other versions present.",
        default=False,
    )


def install_main(args):
    """
    Main function of the 'install' command. Only to be used via the cli.
    :param args: the parsed arguments from the install subparser
    :return:
    """

    # Post process the arguments
    if args.force:
        # uninstall silently
        uninstall(args.target)

    if install(version=args.version, target=args.target):
        print("Installation failed")
        sys.exit(-1)


def determine_platform():
    details = platform.uname()
    if re.match(".*windows.*", details.system, flags=re.IGNORECASE):
        system = "_windows"
    elif re.match(".*darwin.*", details.system, flags=re.IGNORECASE):
        system = "_darwin"
    elif re.match(".*linux.*", details.system, flags=re.IGNORECASE):
        system = "_linux"
    else:
        raise Exception("Unknown system platform")

    if re.match(".*64.*", details.machine):
        arch = "_amd64"
    else:
        arch = "_386"

    return system + arch


def install(version: str = None, target=None):
    """Install main function."""
    version = version or get_latest_version()
    if version.startswith("@"):
        print(f"Version information taken from {version[1:]}")
        version = open(version[1:]).read().strip()

    system = determine_platform()
    file_name = f"databricks_cli_{version}{system}"

    target = target or _DEFAULT_TARGET

    if target.exists():
        raise Exception(f"Target {str(target)} exists. Please uninstall first.")

    if not target.parent.exists():
        raise Exception(
            f"Installation target directory does not exits {str(target.parent)}"
        )

    with open(target, "wb") as t:
        r = requests.get(
            f"{_DATABRICKS_CLI_REPO}/releases/download/v{version}/{file_name}.zip"
        )
        z = zipfile.ZipFile(io.BytesIO(r.content))

        (db_exe_file,) = [n for n in z.namelist() if n.startswith("databricks")]
        t.write(z.read(db_exe_file))

    if not isWin():
        os.chmod(target, 0o755)

    subprocess.run([target, "--version"])



import sys
from types import SimpleNamespace

from spetlrtools.az_databricks_token.AuthLinkOpener import AuthLinkOpener
from spetlrtools.az_databricks_token.server import LocalServer


def get_impersonation_authorization_code(appId: str, tenantId: str):
    # to get the callback, launch a server on localhost
    webServer = LocalServer()
    print("Opening browser, please log in...", file=sys.stderr)

    # open the webbrowser in another thread after two seconds.
    opener = AuthLinkOpener(
        appId=appId, tenantId=tenantId, local_port=webServer.server_port
    )

    # handle one request (blocking)
    webServer.handle_request()

    try:
        result = webServer.last_query
    except AttributeError:
        raise Exception("No callback was received in time. Please re-try.")

    if result.state != opener.state:
        raise Exception("Invalid state field in result.")

    return SimpleNamespace(
        auth_code=result.code, state=opener.state, redirect_uri=opener.redirect_uri
    )



"""
This script is dedicated to the uninstallation of the Databricks CLI tool. Its main features include:
- Setting up a subparser for the 'uninstall' command, allowing the user to specify the target path for uninstallation.
- The main function for the 'uninstall' command, which executes the uninstallation process.
- The core uninstall function, which attempts to remove the Databricks CLI using system commands and
    handles the deletion of the target executable file.

This script is particularly useful for cleanly removing the Databricks CLI from a system, ensuring that
    no residual files are left behind. It utilizes the 'subprocess' module to execute system-level commands
    and 'argparse' for command-line argument parsing.

Note: This script should be run through the command line interface and assumes proper permissions for
    uninstallation tasks.
"""

import argparse
import os
import subprocess
import sys

from spetlrtools.cli_install.defaults import _DEFAULT_TARGET


def setup_uninstall_parser(subparsers):
    """
    Adds a subparser for the command 'install'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "uninstall", description="remove the databricks cli."
    )
    parser.set_defaults(func=uninstall_main)

    parser.add_argument(
        "--target",
        type=str,
        help="path of the installed cli",
        default=_DEFAULT_TARGET,
    )


def uninstall_main(args):
    """
    Main function of the 'uninstall' command. Only to be used via the cli.
    :param args: the parsed arguments from the install subparser
    :return:
    """

    if uninstall(target=args.target):
        print("Uninstallation failed")
        sys.exit(-1)


def uninstall(target=None):
    """Uninstall main function."""

    try:
        subprocess.run(
            ["pip", "uninstall", "databricks-cli", "-y"],
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        "it seems cli v1 was not installed."

    target = target or _DEFAULT_TARGET
    if target.exists():
        os.unlink(target)



import base64
import datetime
import shutil
from dataclasses import dataclass
from pathlib import Path, PosixPath
from tempfile import TemporaryDirectory
from typing.io import BinaryIO

from databricks.sdk import WorkspaceClient
from databricks.sdk.service import workspace

from spetlrtools.test_job.dbcli import DbCli


class StageArea:
    """
    The stage are is either a temp-dir which is cleaned after the run, or, if --dry-run is used,
    the stage area is a local folder whose contents can be inspected for debug purposes.

    In the stage area, the main.py, the libraries and the job.json are collected before submission
    of the job.
    """

    def __init__(self, dry_run=False):
        self.dry_run = dry_run
        self._tmpdir = None

    def __enter__(self) -> str:
        if self.dry_run:
            self._stage_area = "./spetlr-test-job-working-area"
            print(f"Working in static stage area {self._stage_area}.")
            Path(self._stage_area).mkdir(parents=True, exist_ok=False)
            return self._stage_area
        else:
            self._tmpdir = TemporaryDirectory()
            return self._tmpdir.__enter__()

    def __exit__(self, exc_type, exc_val, traceback):
        if self.dry_run:
            print(f"Remember to delete the stage area {self._stage_area}.")
        else:
            return self._tmpdir.__exit__(exc_type, exc_val, traceback)


class RemoteLocation:
    """This class has two functions:
    - decide on a name for the remote location where we upload all the job details
    - present the databricks-internal path to that location
    - execute the upload
    """

    _dbwsc: WorkspaceClient

    date = datetime.datetime.now().isoformat()  # mockable as class member

    def __init__(self, stage_area: str):
        self.stage_area = Path(stage_area)
        self._dbwsc = DbCli().get_client()
        self.me = self._dbwsc.current_user.me().user_name
        self.remote_home_to_base = ""
        self.remote_home = PosixPath()

    def add_local_path(self, source: str, dir: str = None) -> str:
        """Add a source file to the target work area under a certain directory.
        The file is staged. Upload only happens at the end when we call .upload()"""
        source = Path(source)

        stage_dir = self.stage_area / self.remote_home_to_base
        if dir is None:
            target_part = Path(self.remote_home_to_base) / source.parts[-1]
        else:
            stage_dir = stage_dir / dir
            target_part = Path(self.remote_home_to_base) / dir / source.parts[-1]

        stage_dir.mkdir(parents=True, exist_ok=True)

        shutil.copy(source, self.stage_area / target_part)
        return str(self.remote_home / target_part)

    @dataclass
    class FileRef:
        remote: str
        local: str

    def _mkdirs(self, path: str):
        """Depending on the selected remote, use the dbfs or the workspace api"""
        raise NotImplementedError()

    def _upload_object(self, path: str, f: BinaryIO):
        """Depending on the selected remote, use the dbfs or the workspace api"""
        raise NotImplementedError()

    def new_local_file(self, name: str) -> FileRef:
        """Add a new file to the target work area under a certain directory.
        The directory is made to exist. The file should be written to the local
        path which is the second return value.
        Upload only happens at the end when we call .upload()"""

        local_source = Path(self.stage_area) / self.remote_home_to_base / name
        local_source.parent.mkdir(exist_ok=True, parents=True)

        remote_target = self.remote_home / self.remote_home_to_base / name

        return self.FileRef(remote=str(remote_target), local=str(local_source))

    def remote_base(self) -> str:
        """The full path of the work area once it has been uploaded to databricks."""
        raise NotImplementedError()

    def upload(self, dry_run=False):
        """Upload the staging area to databricks, either under dbfs root or under the workspace home folder."""
        if dry_run:
            print("Not uploading test job folder - Action skipped for dry-run.")
        else:
            print("Now uploading test job folder")
            self._upload_dir(self.remote_home, self.stage_area)

    def _upload_dir(self, remote, local):
        """recursively upload directories and files."""
        for obj in Path(local).iterdir():
            if obj.is_dir():
                new_remote = f"{remote}/{obj.name}"
                new_local = str(obj)
                self._mkdirs(f"{remote}/{obj.name}")
                self._upload_dir(new_remote, new_local)
            else:
                with open(obj, "rb") as f:
                    self._upload_object(path=f"{remote}/{obj.name}", f=f)


class WorkspaceLocation(RemoteLocation):
    """Use the Workspace API for remote files."""

    def __init__(self, stage_area: str):
        super().__init__(stage_area)
        self.remote_home_to_base = f".spetlr/test/{self.date}"
        self.remote_home = PosixPath(f"/Workspace/Users/{self.me}")

    def remote_base(self) -> str:
        return str(self.remote_home / self.remote_home_to_base)

    def _mkdirs(self, path: str):
        self._dbwsc.workspace.mkdirs(path)

    def _upload_object(self, path: str, f: BinaryIO):
        content = base64.b64encode(f.read()).decode()
        self._dbwsc.workspace.import_(
            path=path, content=content, format=workspace.ImportFormat.AUTO
        )


class DbfsLocation(RemoteLocation):
    """Use the DBFS API for remote files."""

    def __init__(self, stage_area: str):
        super().__init__(stage_area)
        self.date = self.date.replace(":", ".")
        self.remote_home_to_base = f"spetlr/test/{self.me}/{self.date}"
        self.remote_home = PosixPath("dbfs:/")

    def remote_base(self) -> str:
        return str(self.remote_home / self.remote_home_to_base)

    def _mkdirs(self, path: str):
        self._dbwsc.dbfs.mkdirs(path)

    def _upload_object(self, path: str, f: BinaryIO):
        self._dbwsc.dbfs.upload(path=path, src=f)



from datetime import datetime, timedelta, timezone
from typing import Union

import dateparser as dateparser


class TimeSequence:
    def __init__(
        self,
        start: Union[datetime, str] = datetime.now(timezone.utc),
        delta_seconds=100,
    ):
        if isinstance(start, datetime):
            self.last = start
        else:
            self.last = dateparser.parse(start).astimezone(timezone.utc)
        self.delta = timedelta(seconds=delta_seconds)

    def reverse(self, steps: int = 1):
        """produce descending timestamps so that first call is latest."""
        return self.next(-1 * steps)

    def next(self, steps: int = 1):
        """produce ascending timestamps so that last call is latest."""
        self.last = self.last + steps * self.delta
        return self.last

    def __next__(self):
        return self.next()



import argparse

from spetlrtools.test_job.fetch import setup_fetch_parser
from spetlrtools.test_job.submit import setup_submit_parser


def main():
    """Cli main function. Not for use from python."""
    parser = argparse.ArgumentParser(
        description="Run Test Cases on databricks cluster."
    )

    subparsers = parser.add_subparsers(required=True, dest="command")

    setup_submit_parser(subparsers)
    setup_fetch_parser(subparsers)

    args = parser.parse_args()
    args.func(args)



from databricks.sdk.service import jobs

from spetlrtools.test_job.dbcli import DbCli


class RunDetails:
    """Object representing the details of a job run"""

    details: jobs.Run

    def __init__(self, run_id: int):
        self.run_id = run_id
        self._db = DbCli()
        self.refresh()
        print(f"Job details: {self.details.run_page_url}")

    def refresh(self):
        """refresh the internal details by querying databricks"""
        self.details = self._db.get_run(self.run_id)

    def cancel(self):
        """Cancel the run on databricks."""
        self._db.cancel_run(self.run_id)

    def get_stdout(self, task_key: str) -> str:
        """Return the driver stdout from the cluster logs."""
        self.refresh()

        print(f"Getting stdout for {task_key}")
        task: jobs.RunTask

        # get latest attempt for all tasks of this key
        # this allows fetch to succeed in case a "repair" fixed a transient task failure
        tasks_for_key = (t for t in self.details.tasks if t.task_key == task_key)
        # sort by negative attempt number
        task = sorted(tasks_for_key, key=lambda t: -t.attempt_number)[0]

        task_id = task.run_id
        output = self._db.get_run_output(task_id)
        return output.logs or task.status.termination_details.message



"""
- find the test job (by ID or discover by tag)
- get job status:
- tasks: pending: 0 running: 0 success: 0 failed: 0
- any time a task finishes, print the log
- add a fail_fast so that any time a task fails, job is cancelled
"""

import argparse
import json
import sys
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import IO, List, Optional

from databricks.sdk.service import jobs

from spetlrtools.test_job.RunDetails import RunDetails


def setup_fetch_parser(subparsers):
    """
    Adds a subparser for the command 'fetch'.
    :param subparsers: must be the object returned by ArgumentParser().add_subparsers()
    :return:
    """
    parser: argparse.ArgumentParser = subparsers.add_parser(
        "fetch", description="Return test run result."
    )
    parser.set_defaults(func=fetch_main)

    # cluster argument pair
    runid_config = parser.add_mutually_exclusive_group(required=True)
    runid_config.add_argument(
        "--runid",
        type=int,
        help="Run ID of the test job",
        default=None,
    )
    runid_config.add_argument(
        "--runid-json",
        type=argparse.FileType("r"),
        help="File with JSON document describing the Run ID of the test job.",
        default=None,
    )

    parser.add_argument(
        "--stdout",
        type=argparse.FileType("w"),
        required=False,
        help="Output test stdout to this file.",
        default=None,
    )

    parser.add_argument(
        "--failfast",
        action="store_true",
        help="Stop and cancel job on first failed task.",
    )


def collect_args(args):
    """Post process the arguments of the ."""
    if args.runid is None:
        args.runid = json.load(args.runid_json)["run_id"]

    return args


def fetch_main(args):
    """
    Main function of the 'fetch' command. Only to be used via the cli.
    :param args: the parsed arguments from the fetch subparser
    :return:
    """

    # Post process the arguments
    if args.runid is None:
        args.runid = json.load(args.runid_json)["run_id"]

    if fetch(args.runid, args.stdout, args.failfast):
        print("Run failed")
        sys.exit(-1)


def fetch(run_id: int, stdout_file: IO[str] = None, failfast=False):
    """Fetch main function.
    See the cli help for parameter descriptions and functionality.
    Can be used programmatically."""

    run = RunDetails(run_id)

    last_state = None
    stdouts = {}
    while True:
        state = MultiTaskState.fromRun(run.details)
        # state = MultiTaskState.fromJson(run.details.as_dict())
        if last_state is None or state != last_state:
            last_state = state
            state.print_status()

        if failfast:
            if any(task.ended and not task.success for task in state.tasks):
                break

        for task in state.tasks:
            if task.ended and task.task_key not in stdouts:
                out = run.get_stdout(task.task_key)
                if stdout_file is None:
                    print(out)
                stdouts[task.task_key] = out

        if state.overall.ended:
            break
        time.sleep(5)
        run.refresh()

    if stdout_file is not None:
        for k, v in stdouts.items():
            stdout_file.write("=" * 50 + f"\nTask Output from {k}\n" + "=" * 50 + "\n")
            stdout_file.write(v)

    if last_state.overall.success:
        print("Run result SUCCESS!")
        return 0
    else:
        if not (last_state.overall.ended):
            print("A task failed. Cancelling test run.")
            run.cancel()
        return 1


def enumNameOrNone(obj):
    if obj is None:
        return ""
    else:
        return obj.name


@dataclass
class TaskState:
    """The result state of any workflow or task"""

    task_key: str
    life_cycle_state: str
    result_state: str
    end_time: int

    @property
    def ended(self):
        """Has the workflow or task ended?"""
        return self.end_time != 0

    @property
    def result(self):
        """String representing the state of the workflow or task."""
        if not self.ended:
            return self.life_cycle_state
        else:
            return self.result_state

    @property
    def success(self):
        """Has the workflow or task succeeded?"""
        return self.result_state.upper() == "SUCCESS"

    @classmethod
    def fromTask(cls, task: jobs.RunTask):
        """Create the result state of the workflow or task from the json object
        returned by the databricks api."""

        return cls(
            task_key=task.task_key,
            life_cycle_state=enumNameOrNone(task.state.life_cycle_state),
            result_state=enumNameOrNone(task.state.result_state),
            end_time=task.end_time,
        )

    @classmethod
    def fromRun(cls, run: jobs.Run):
        """Create the result state of the workflow or task from the json object
        returned by the databricks api."""

        return cls(
            task_key=run.run_name,
            life_cycle_state=enumNameOrNone(run.state.life_cycle_state),
            result_state=enumNameOrNone(run.state.result_state),
            end_time=run.end_time,
        )


@dataclass
class MultiTaskState:
    """Result state of a multiTask workflow"""

    overall: Optional[TaskState]
    tasks: List[TaskState]

    @classmethod
    def fromRun(cls, run: jobs.Run):
        """Create the Result state of a multiTask workflow from the json object returned
        by the databricks api."""

        return cls(
            overall=TaskState.fromRun(run),
            tasks=[TaskState.fromTask(task) for task in run.tasks],
        )

    def accumulate(self):
        """return a dictionary of {'STATE STRING':counts} where counts represents the
        number of sub-tasks that share the same run state."""
        counts = defaultdict(int)
        for task in self.tasks:
            counts[task.result] += 1
        return counts

    def print_status(self):
        """Print the overall result state represented by this object."""
        print(
            "Overall state:",
            self.overall.result,
            "| Task states:",
            " | ".join(f"{k}: {v}" for k, v in self.accumulate().items()),
        )
